// -*- coding: iso-8859-1-unix -*-

/**
* API do OpenBus SDK C++
* \file openbus/connection.hpp
* 
*/

#ifndef TECGRAF_SDK_OPENBUS_OPENBUS_CONNECTION_HPP
#define TECGRAF_SDK_OPENBUS_OPENBUS_CONNECTION_HPP

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
#include "scsC.h"
#pragma clang diagnostic pop
#include "openbus/idl.hpp"
#include "openbus/detail/decl.hpp"
#include "openbus/detail/interceptors/orb_initializer.hpp"
#ifndef TECGRAF_SDK_OPENBUS_OPENBUS_DETAIL_LRUCACHE_HPP
#define TECGRAF_SDK_OPENBUS_OPENBUS_DETAIL_LRUCACHE_HPP
#include "openbus/detail/lru_cache.hpp"
#endif
#include "openbus/orb_initializer.hpp"
#include "openbus/detail/openssl/private_key.hpp"
#include "openbus/shared_auth_secret.hpp"

#include <boost/array.hpp>
#include <boost/function.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/thread.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>
#include <cstring>

/**
* \brief openbus
*/
namespace openbus 
{  
  class OpenBusContext;
  class LoginCache;
  class RenewLogin;
  class PublicKey;

  namespace interceptors
  {
    struct orb_info;
    struct ServerInterceptor;
    struct ClientInterceptor;
  }
}
 
namespace openbus 
{ 
struct OPENBUS_SDK_DECL WrongBus : public std::exception 
{ 
  const char *what() const throw()
  { 
    return "openbus::WrongBus";
  }
};

struct OPENBUS_SDK_DECL BusChanged : public std::exception 
{ 
  const char *what() const throw()
  { 
    return "openbus::BusChanged";
  }
};

struct OPENBUS_SDK_DECL AlreadyLoggedIn : public std::exception 
{ 
  const char *what() const throw()
  { 
    return "openbus::AlreadyLoggedIn"; 
  }
};

struct OPENBUS_SDK_DECL InvalidLoginProcess : public std::exception 
{
  const char *what() const throw()
  { 
    return "openbus::InvalidLoginProcess"; 
  }
};

struct OPENBUS_SDK_DECL InvalidPropertyValue : public std::exception 
{
  InvalidPropertyValue(const std::string &p, const std::string &v) throw()
    : property(p), value(v) 
  { 
  }

  ~InvalidPropertyValue() throw()
  { 
  }

  const char *what() const throw() 
  { 
    return "openbus::InvalidPropertyValue"; 
  }

  const std::string property;
  const std::string value;
};

struct OPENBUS_SDK_DECL InvalidBusAddress : public std::exception 
{
  const char *what() const throw()
  { 
    return "openbus::InvalidBusAddress"; 
  }
};

/**
 * \brief Conexo para acesso identificado a um barramento.
 *
 * Uma conexo  usada para realizar acessos identificados a um barramento.
 * Denominamos esses acessos identificados ao barramento de login. Cada login
 * possui um identificador nico e est sempre associado ao nome de uma
 * entidade que  autenticada no momento do estabelecimento do login.
 * H basicamente trs formas de autenticao de entidade disponveis:
 * - Por Senha: veja a operao \ref loginByPassword
 * - Por Certificado de login: veja a operao \ref loginByCertificate
 * - Por Autenticao compartilhada: veja a operao \ref loginBySharedAuth
 *
 * A entidade associada ao login  responsvel por todas as chamadas feitas
 * atravs daquela conexo e essa entidade deve ser levada em considerao
 * pelos servios ofertados no barramento para decidir aceitar ou recusar
 * chamadas.
 *
 *  possvel uma aplicao assumir mltiplas identidades ao acessar um ou mais
 * barramentos criando mltiplas conexes para esses barramentos.
 * 
 *  importante notar que a conexo no  usada diretamente pela aplicao ao
 * realizar ou receber chamadas, pois as chamadas ocorrem usando proxies e
 * servants de um ORB. As conexes que so efetivamente usadas nas chamadas do
 * ORB so definidas atravs do OpenBusContext associado ao ORB.
 */
class OPENBUS_SDK_DECL Connection
  : public boost::enable_shared_from_this<Connection>
{
public:
  /**
   * \brief Callback de login invlido.
   * 
   * Tipo que representa um objeto funo ('function object') a ser chamado
   * quando uma notificao de login invlido  recebida. Caso alguma exceo
   * ocorra durante a execuo do mtodo e no seja tratada, o erro ser
   * capturado pelo interceptador e registrado no log.
   * 
   * O tipo InvalidLoginCallback  um typedef de boost::function. Para
   * documentao dessa biblioteca acesse
   * http://www.boost.org/doc/libs/1_47_0/doc/html/function.html
   *
   * \param conn Conexo que recebeu a notificao de login invlido.
   * \param login Informaes do login que se tornou invlido.
   */
  typedef boost::function<void (Connection &conn,
                                idl::access::LoginInfo login)> 
    InvalidLoginCallback;
  
  /**
  * Efetua login no barramento como uma entidade usando autenticao por senha.
  * 
  * A autenticao por senha  validada usando um dos validadores de senha
  * definidos pelo adminsitrador do barramento.
  *
  * \param entity Identificador da entidade a ser autenticada.
  * \param password Senha de autenticao no barramento da entidade.
	* \param domain Identificador do domnio de autenticao.
  * 
  * \throw AlreadyLoggedIn A conexo j est logada.
  * \throw idl::access::AccessDenied Senha fornecida para autenticao da 
  *        entidade no foi validada pelo barramento.
  * \throw TooManyAttempts A autenticao foi recusada por um nmero 
  *        excessivo de tentativas invlidas de login por senha.
  * \throw idl::core::ServiceFailure Ocorreu uma falha interna nos servios 
  *        do barramento que impediu a autenticao da conexo.
	* \throw idl::access::UnknownDomain O domnio de autenticao no  conhecido.
  * \throw idl::access::WrongEncoding A autenticao falhou, pois a senha no 
  *        foi codificada corretamente com a chave pblica do barramento.
  * \throw CORBA::Exception
  */
  void loginByPassword(
    const std::string &entity,
    const std::string &password,
    const std::string &domain = "");
  
 /**
  * \brief Efetua login de uma entidade usando autenticao por certificado.
  * 
  * A autenticao por certificado  validada usando um certificado de login
  * registrado pelo administrador do barramento.
  * 
  * \param entity Identificador da entidade a ser conectada.
  * \param key Chave privada da entidade utilizada na autenticao. 
  *             uma precondio um EVP_PKEY vlido e no nulo.
  * 
  * \throw AlreadyLoggedIn A conexo j est autenticada.
  * \throw idl::access::AccessDenied A chave privada fornecida no corresponde 
  &        ao certificado da entidade registrado no barramento indicado.
  * \throw idl::access::MissingCertificate No h certificado para essa 
  *        entidade registrado no barramento indicado.
  * \throw idl::access::WrongEncoding A autenticao falhou, pois a senha no 
  *        foi codificada corretamente com a chave pblica do barramento.
  * \throw idl::core::ServiceFailure Ocorreu uma falha interna nos servios 
  *        do barramento que impediu a autenticao da conexo.
  * \throw CORBA::Exception
  */
  void loginByCertificate(
    const std::string &entity,
    EVP_PKEY *key);
  
  /**
  * \brief Inicia o processo de login por autenticao compartilhada.
  * 
  * A autenticao compartilhada permite criar um novo login compartilhando a
  * mesma autenticao do login atual da conexo. Portanto essa operao s pode
  * ser chamada enquanto a conexo estiver autenticada, caso contrrio a exceo
  * de sistema CORBA::NO_PERMISSION{NoLogin}  lanada. As informaes
  * fornecidas por essa operao devem ser passadas para a operao
  * \ref loginBySharedAuth para concluso do processo de login por autenticao
  * compartilhada. Isso deve ser feito dentro do tempo de lease definido pelo
  * administrador do barramento. Caso contrrio essas informaes se tornam
  * invlidas e no podem mais ser utilizadas para criar um login.
  * 
  * \return Segredo a ser fornecido na concluso do processo de login.
  *
  * \throw idl::core::ServiceFailure 
  *        Ocorreu uma falha interna nos servios do barramento que impediu 
  *        o estabelecimento da conexo.
  * \throw CORBA::Exception
  */
  SharedAuthSecret startSharedAuth();
  
  /**
  * \brief Efetua login de uma entidade usando autenticao compartilhada.
  * 
  * A autenticao compartilhada  feita a partir de um segredo obtido
  * atravs da operao 'startSharedAuth' de uma conexo autenticada.
  * 
  * \param secret Segredo a ser fornecido na concluso do processo de login.
  * 
  * \throw WrongBus O segredo no pertence ao barramento contactado.
  * \throw InvalidLoginProcess A tentativa de login associada ao segredo
  *        informado  invlido, por exemplo depois do segredo ser
  *        cancelado, ter expirado, ou j ter sido utilizado.
  * \throw AlreadyLoggedIn A conexo j est autenticada.
  * \throw AccessDenied O segredo fornecido no corresponde ao esperado
  *        pelo barramento.
  * \throw idl::access::WrongEncoding A autenticao falhou, pois a senha 
  *        no foi codificada corretamente com a chave pblica do barramento.
  * \throw ServiceFailure Ocorreu uma falha interna nos servios do
  *        barramento que impediu a autenticao da conexo.
  * \throw CORBA::Exception
  */
  void loginBySharedAuth(const SharedAuthSecret &secret);
  
  /**
	 * \brief Efetua logout da conexo, tornando o login atual invlido.
	 * 
	 * Aps a chamada a essa operao a conexo fica desautenticada, implicando que
	 * qualquer chamada realizada pelo ORB usando essa conexo resultar numa
	 * exceo de sistema 'CORBA::NO_PERMISSION{NoLogin}' e chamadas recebidas
	 * por esse ORB sero respondidas com a exceo
	 * 'CORBA::NO_PERMISSION{UnknownBus}' indicando que no foi possvel
	 * validar a chamada pois a conexo est temporariamente desautenticada.
	 * 
	 * \return Verdadeiro se o processo de logout for concludo com xito e 
	 *         falso se no for possvel invalidar o login atual.
	 */
  bool logout();
  	
  /**
   * \brief Callback a ser chamada quando o login atual se tornar invlido.
   *
   * Esse atributo  utilizado para definir um objeto funo (function object)
   * que implementa uma interface de callback a ser chamada sempre que a conexo
   * receber uma notificao de que o seu login est invlido. Essas
   * notificaes ocorrem durante chamadas realizadas ou recebidas pelo
   * barramento usando essa conexo. Um login pode se tornar invlido caso o
   * administrador explicitamente o torne invlido ou caso a thread interna de
   * renovao de login no seja capaz de renovar o lease do login a tempo. Caso
   * esse atributo seja um InvalidLoginCallback 'default-constructed', nenhum
   * objeto de callback  chamado na ocorrncia desse evento.
   *
   * Durante a execuo dessa callback um novo login pode ser restabelecido.
   * Neste caso, a chamada do barramento que recebeu a notificao de login
   * invlido  refeita usando o novo login, caso contrrio, a chamada original
   * lana a exceo de de sistema 'CORBA::NO_PERMISSION{NoLogin}'.
   * 
   * O tipo InvalidLoginCallback  um typedef de boost::function. Para
   * documentao dessa biblioteca acesse
   * http://www.boost.org/doc/libs/1_47_0/doc/html/function.html
   */
  void onInvalidLogin(InvalidLoginCallback p);
  
  /**
   * \brief Retorna a callback configurada para ser chamada quando o login atual
   * se torna invlido.
   */
  InvalidLoginCallback onInvalidLogin() const;
  
  /**
   * \brief Informaes do login dessa conexo ou nulo se a conexo no est
   * autenticada, ou seja, no tem um login vlido no barramento.
   */
  const idl::access::LoginInfo *login() const;
  
  /**
   * \brief Identificador do barramento ao qual essa conexo se refere.
   */
  const std::string busid() const;
  
  ~Connection();  
private:
  /**
  * Connection deve ser adquirido atraves de 
  * OpenBusContext::connectByReference ou OpenBusContext::connectByReference.
  */
  Connection(CORBA::Object_var,
             CORBA::ORB_ptr, 
             interceptors::ORBInitializer *,
             OpenBusContext &, 
             EVP_PKEY *access_key,
             bool legacy_support);

  /**
  * Connection deve ser adquirido atraves de
  * OpenBusContext::connectByAddress ou OpenBusContext::connectByReference.
  */
  Connection(const std::string host,
             const unsigned short port,
             CORBA::ORB_ptr, 
             interceptors::ORBInitializer *,
             OpenBusContext &, 
             EVP_PKEY *access_key,
             bool legacy_support);

  Connection(const Connection &);
  Connection &operator=(const Connection &);

  static void renewLogin(
    boost::weak_ptr<Connection> conn,
    idl::access::AccessControl_ptr acs, 
    OpenBusContext &ctx,
    idl::access::ValidityTime t);

  void login(idl::access::LoginInfo &loginInfo, 
             idl::access::ValidityTime validityTime);

  void checkBusid() const;
  bool _logout(bool local = true);
  CORBA::ORB_ptr orb() const 
  { 
    return _orb; 
  }

  idl::access::LoginRegistry_var login_registry() const 
  { 
    return _login_registry; 
  }

  idl::access::AccessControl_var access_control() const 
  { 
    return _access_control; 
  }

  const idl::access::LoginInfo *_login() const 
  { 
    boost::lock_guard<boost::mutex> lock(_mutex);;
    return _loginInfo.get(); 
  }

  idl::offers::OfferRegistry_var getOfferRegistry() const
  { 
    return _offer_registry;
  }

  idl::access::LoginRegistry_var getLoginRegistry() const
  {
    return _login_registry;
  }

  void init();

  idl::access::LoginInfo get_login();

  /* Variaveis que sao modificadas somente no construtor. */
  OpenBusContext &_openbusContext;
  PrivateKey _key;
  idl::access::AccessControl_var _access_control;
  idl::access::LoginRegistry_var _login_registry;
  idl::offers::OfferRegistry_var _offer_registry;
  boost::scoped_ptr<LoginCache> _loginCache;
  std::string _busid;
  boost::scoped_ptr<PublicKey> _buskey;
  bool _legacy_support;
  idl::legacy::access::AccessControl_var _legacy_access_control;
  idl::legacy_support::LegacyConverter_var _legacy_converter;
  /**/

  CORBA::Object_var _component_ref;
  scs::core::IComponent_var _iComponent;
  const std::string _host;
  const unsigned short _port;
  interceptors::ORBInitializer * _orb_init;
  CORBA::ORB_ptr _orb;
  boost::thread _renewLogin;
  mutable boost::mutex _mutex;
  boost::scoped_ptr<idl::access::LoginInfo> _loginInfo, _invalid_login;
  InvalidLoginCallback _onInvalidLogin;
  
  enum State 
  {
    LOGGED,
    UNLOGGED,
    INVALID
  } _state;
  
    
  struct SecretSession 
  {
    SecretSession()
      : id(0)
      , ticket(0)
      , is_legacy(false)
    {
      secret.fill(0);
    }
    CORBA::ULong id;
    std::string remote_id, entity;
    boost::array<unsigned char, secret_size> secret;
    CORBA::ULong ticket;
    bool is_legacy;
    friend bool operator==(const SecretSession &lhs, const SecretSession &rhs);
    friend bool operator!=(const SecretSession &lhs, const SecretSession &rhs);
  };
  typedef LRUCache<hash_value, std::string> profile2login_LRUCache;
  profile2login_LRUCache _profile2login;
  LRUCache<std::string, SecretSession> _login2session;

  friend struct openbus::interceptors::ServerInterceptor;
  friend struct openbus::interceptors::ClientInterceptor;
  friend class openbus::OpenBusContext;
  friend bool operator==(const SecretSession &lhs, const SecretSession &rhs);
  friend bool operator!=(const SecretSession &lhs, const SecretSession &rhs);
};

inline bool operator==(const Connection::SecretSession &lhs, 
                const Connection::SecretSession &rhs)
{
  return lhs.id == rhs.id
    && lhs.remote_id == rhs.remote_id
    && lhs.secret == rhs.secret
    && lhs.ticket == rhs.ticket;
}

inline bool operator!=(const Connection::SecretSession &lhs, 
                const Connection::SecretSession &rhs)
{
  return !(lhs == rhs);
}

}

#endif
