// -*- coding: iso-8859-1-unix -*-
/**
  \mainpage API do OpenBus SDK C++
  
  Este documento apresenta uma API C++ para o acesso ao barramento
  OpenBus. O contedo aqui apresentado foi extrado automaticamente
  dos fontes da biblioteca. 

  O ponto de entrada da interface  a classe \ref
  openbus::ORBInitializer, que retorna um invlucro para o
  ORB. Atravs do ORB,  possvel obter a instncia para \ref
  openbus::OpenBusContext, que permite controlar o contexto das
  chamadas de um ORB para acessar informaes que identificam essas
  chamadas em barramentos OpenBus.

  <a href="https://jira.tecgraf.puc-rio.br/confluence/pages/viewpage.action?pageId=67600726">Manual de uso do ncleo do OpenBus 2.1</a>

  <a href="https://jira.tecgraf.puc-rio.br/confluence/display/OPENBUS021/CORE">Download do barramento OpenBus 2.1</a>
 
  \file openbus/orb_initializer.hpp
*/

#ifndef TECGRAF_SDK_OPENBUS_OPENBUS_ORB_INITIALIZER_HPP
#define TECGRAF_SDK_OPENBUS_OPENBUS_ORB_INITIALIZER_HPP

#include "openbus/detail/decl.hpp"

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
#include <tao/ORB.h>
#pragma clang diagnostic pop
#include <boost/shared_ptr.hpp>

/**
* \brief Namespace para a biblioteca de acesso.
*/
namespace openbus {

/**
 * \class orb_ctx
 * \brief 
 * Representa um ORB atravs de um invlucro que assume a 
 * propriedade(ownership) sobre o ORB. O destrutor do invlucro chama
 * CORBA::ORB::destroy(). A instncia do ORB pode ser obtida atravs do
 * mtodo \ref orb.
 *
 */
class OPENBUS_SDK_DECL orb_ctx
{
public:
  /**
   * \brief Construtor que recebe um 'smart pointer' de CORBA 
   * para a instncia do ORB(CORBA::ORB_var).
   */
  orb_ctx(CORBA::ORB_var orb);

  /**
   * \brief Destrutor que chama orb::destroy().
   */
  ~orb_ctx();

  /**
   * \brief Compartilha a instncia do ORB.
   *
   * Retorna um 'smart pointer' para a instncia do ORB que  
   * referenciado internamente.
   * 
   * \return CORBA::ORB_var
   */
  CORBA::ORB_var orb() const
  {
    return orb_;
  }
private:
  orb_ctx &operator=(const orb_ctx &);
  CORBA::ORB_var orb_;
};

/**
 * \brief Inicializa um ORB utilizado exclusivamente para chamadas
 *        atravs de barramentos OpenBus.
 *
 * Este ORB no pode ser utilizado para fazer chamadas CORBA
 * ordinrias sem o controle de acesso do OpenBus que permite
 * identificao da origem das chamadas. Esse controle de acesso 
 * feito atravs de conexes, que so obtidas e manipuladas atravs de
 * um \ref OpenBusContext. O ORB possui um \ref OpenBusContext
 * associado, que pode ser obitido atravs da chamada: 
 * \code
 * CORBA::ORB::resolve_initial_reference("OpenBusContext") 
 * \endcode
 *
 * O ORB  inicializado da mesma forma feita pela operao
 * 'CORBA::ORB_init' definida pelo padro CORBA. 
 * 
 * Chamadas realizadas e recebidas atravs deste ORB so interceptadas
 * pela biblioteca de acesso do OpenBus e podem lanar excees de
 * sistema de CORBA definidas pelo OpenBus. A seguir so apresentadas
 * essas excees: 
 *  - CORBA::NO_PERMISSION{NoLogin}: Nenhuma conexo
 *    "Requester" com login vlido est associada ao contexto atual, ou
 *     seja, a conexo "Requester" corrente est desautenticada.  
 *  - CORBA::NO_PERMISSION{InvalidChain}: A cadeia de chamadas associada
 *    ao contexto atual no  compatvel com o login da conexo
 *    "Requester" desse mesmo contexto. Isso ocorre pois no  possvel
 *    fazer chamadas dentro de uma cadeia recebida por uma conexo com um
 *    login diferente.  
 *  - CORBA::NO_PERMISSION{UnknownBus}: O ORB remoto
 *    que recebeu a chamada indicou que no possui uma conexo com login
 *    vlido no barramento atravs do qual a chamada foi realizada,
 *    portanto no  capaz de validar a chamada para que esta seja
 *    processada.  
 *  - CORBA::NO_PERMISSION{UnverifiedLogin}: O ORB remoto
 *    que recebeu a chamada indicou que no  capaz de validar a chamada
 *    para que esta seja processada. Isso indica que o lado remoto tem
 *    problemas de acesso aos servios ncleo do barramento.  
 *  - CORBA::NO_PERMISSION{InvalidRemote}: O ORB remoto que recebeu a
 *    chamada no est se comportando de acordo com o protocolo OpenBus
 *    2.1, o que indica que est mal implementado e tipicamente
 *    representa um bug no servidor sendo chamado ou um erro de
 *    implantao do barramento.
 * 
 * \param argc Nmero de parmetros usados na inicializao do ORB.
 * \param argv Parmetros usados na inicializao do ORB.
 * 
 * \throw CORBA::Exception
 *
 * \return Um 'smart pointer' do tipo \ref std::auto_ptr para um 
 * invlucro que contm o ORB inicializado.
 */
  OPENBUS_SDK_DECL std::auto_ptr<orb_ctx>
  ORBInitializer(int &argc, char **argv);
}

#endif
