// -*- coding: iso-8859-1-unix -*-
/**
* API do OpenBus SDK C++
* \file openbus/openbus_context.hpp
*/

#ifndef TECGRAF_SDK_OPENBUS_OPENBUS_CONTEXT_HPP
#define TECGRAF_SDK_OPENBUS_OPENBUS_CONTEXT_HPP

#include "openbus/idl.hpp"
#include "openbus/detail/decl.hpp"
#include "openbus/orb_initializer.hpp"
#include "openbus/shared_auth_secret.hpp"
#include "openbus/caller_chain.hpp"

#include <tao/LocalObject.h>
#include <openssl/evp.h>
#include <boost/function.hpp>
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wunused-local-typedef"
#include <boost/thread.hpp>
#pragma clang diagnostic pop
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/parameter.hpp>
#include <boost/uuid/uuid.hpp>
#include <string>
#include <map>
#include <set>

namespace openbus 
{
  class Connection;
  namespace interceptors
  {
    struct ServerInterceptor;
    struct ClientInterceptor;
  }
}

/**
* \brief openbus
*/
namespace openbus 
{

struct OPENBUS_SDK_DECL InvalidEncodedStream : public std::exception
{
  InvalidEncodedStream();
  InvalidEncodedStream(const std::string &msg);
  ~InvalidEncodedStream() throw();
  const char *what() const throw()
  {
    return msg_.c_str();
  }
private:
  std::string msg_;
};  

BOOST_PARAMETER_NAME(host)
BOOST_PARAMETER_NAME(port)
BOOST_PARAMETER_NAME(access_key)
BOOST_PARAMETER_NAME(legacy_support)
BOOST_PARAMETER_NAME(reference)

/**
 * \class OpenBusContext
 * \brief Permite controlar o contexto das chamadas de um ORB para acessar
 *        informaes que identificam essas chamadas em barramentos OpenBus.
 *
 * O contexto de uma chamada pode ser definido pela linha de execuo atual
 * do programa em que executa uma chamada, o que pode ser a thread em execuo
 * ou mais comumente o 'CORBA::PICurrent' do padro CORBA. As informaes
 * acessveis atravs do OpenBusContext se referem basicamente 
 * identificao da origem das chamadas, ou seja, nome das entidades que
 * autenticaram os acessos ao barramento que originaram as chamadas.
 * 
 * A identifcao de chamadas no barramento  controlada atravs do
 * OpenBusContext atravs da manipulao de duas abstraes representadas
 * pelas seguintes interfaces:
 * - Connection: Representa um acesso ao barramento, que  usado tanto para
 *   fazer chamadas como para receber chamadas atravs do barramento. Para
 *   tanto a conexo precisa estar autenticada, ou seja, logada. Cada chamada
 *   feita atravs do ORB  enviada com as informaes do login da conexo
 *   associada ao contexto em que a chamada foi realizada. Cada chamada
 *   recebida tambm deve vir atravs de uma conexo logada, que deve ser o
 *   mesmo login com que chamadas aninhadas a essa chamada original devem ser
 *   feitas.
 * - CallerChain: Representa a identicao de todos os acessos ao barramento que
 *   originaram uma chamada recebida. Sempre que uma chamada  recebida e
 *   executada,  possvel obter um CallerChain atravs do qual  possvel
 *   inspecionar as informaes de acesso que originaram a chamada recebida.
 */
class OPENBUS_SDK_DECL OpenBusContext : public CORBA::LocalObject 
{
public:
  /**
	 * \brief Callback de despacho de chamadas.
	 * 
	 * Mtodo a ser implementado pelo objeto de callback a ser chamado quando
	 * uma chamada proveniente de um barramento  recebida. Esse mtodo  chamado
	 * para determinar a conexo a ser utilizada na validao de cada chamada
	 * recebida. Se a conexo informada no estiver conectada ao mesmo barramento
	 * indicado pelo parmetro 'busid', a chamada provavelmente ser recusada com
	 * um CORBA::NO_PERMISSION{InvalidLogin} pelo fato do login provavelmente no
	 * ser vlido no barramento da conexo. Como resultado disso o cliente da
	 * chamada poder indicar que o servidor no est implementado corretamente e
	 * lanar a exceo CORBA::NO_PERMISSION{InvalidRemote}. Caso alguma exceo
	 * ocorra durante a execuo do mtodo e no seja tratada, o erro ser
	 * capturado pelo interceptador e registrado no log.
	 * 
	 * \param bus_ctx Gerenciador de contexto do ORB que recebeu a chamada.
	 * \param bus_id Identificao do barramento atravs do qual a chamada foi
	 *                  feita.
	 * \param login_id Informaes do login do cliente da chamada.
	 * \param operation Nome da operao sendo chamada.
	 *
	 * \return Conexo a ser utilizada para receber a chamada.
	 */
  typedef boost::function<
    boost::shared_ptr<Connection> (
      OpenBusContext &bus_ctx,
      const std::string bus_id, 
      const std::string login_id,
      const std::string operation)>
    CallDispatchCallback;

  /**
	 * \brief Callback a ser chamada para determinar a conexo a ser utilizada
	 *        para receber cada chamada.
	 *
	 * Definio de um objeto que implementa uma interface de callback a
	 * ser chamada sempre que a conexo receber uma chamada do
	 * barramento. Essa callback deve devolver a conexo a ser utilizada
	 * para receber a chamada. A conexo utilizada para receber a
	 * chamada ser a nica conexo atravs da qual novas chamadas
	 * aninhadas  chamada recebida podero ser feitas (veja a operao
	 * \ref joinChain).
	 *
	 * Se o objeto de callback for definido como nulo ou devolver nulo, a
	 * conexo padro  utilizada para receber a chamada, caso esta esteja
	 * definida.
	 *
	 * Caso esse atributo seja nulo, nenhum objeto de callback  chamado.
	 */
  void onCallDispatch(CallDispatchCallback);

  /**
   * \brief Retorna a callback a ser chamada para determinar a conexo
   *        a ser utilizada para receber cada chamada
   *
   * Veja a operao \ref onCallDispatch e o tipo \ref CallDispatchCallback 
   * para maiores informaes.  
   */
  CallDispatchCallback onCallDispatch() const;

  /**
	 * \brief Cria uma conexo para um barramento indicado por uma referncia
	 *        CORBA.
   * 
	 * O barramento  indicado por uma referncia CORBA a um componente
	 * SCS que representa os servios ncleo do barramento. Essa funo
	 * deve ser utilizada ao invs da connectByAddress para permitir o
	 * uso de SSL nas comunicaes com o ncleo do barramento.
   * Este mtodo aceita argumentos no formato Boost Parameter.
   *
	 * \param _reference Referncia CORBA a um componente SCS que representa os
	 *            servios ncleo do barramento.
   * \param _access_key (opcional) chave de acesso a ser utilizada
   *            internamente para a gerao de credenciais que identificam 
   *            as chamadas atravs do barramento. A chave deve ser uma 
   *            chave privada RSA de 2048 bits (256 bytes). O parmetro 
   *            do tipo EVP_PKEY* e na sua ausncia uma chave de acesso 
   *            gerada automaticamente.
   * \param _legacy_support (opcional) boleano que desabilita o suporte
   *            a chamadas usando o protocolo OpenBus 2.0. Por padro o suporte
   *            est habilitado
   *
   * Exemplo de chamadas sem os parmetros opcionais:
   * \code
   * std::stringstream corbaloc;
   * corbaloc << "corbaloc::" 
   *          << "localhost" << ":" << 2089
   *          << "/" << tecgraf::openbus::core::v2_1::BusObjectKey;
   * CORBA::Object_var
   *   ref(orb_ctx->orb()->string_to_object(corbaloc.str().c_str()));
   * connectByReference(ref.in());
   * \endcode
   *
   * Exemplo de chamada com os parmetros opcionais:
   * \code
   * std::stringstream corbaloc;
   * corbaloc << "corbaloc::" << "localhost" << ":" << 2089
              << "/" << tecgraf::openbus::core::v2_1::BusObjectKey;
   * CORBA::Object_var
   *   ref(orb_ctx->orb()->string_to_object(corbaloc.str().c_str()));
   * connectByReference(ref.in(),
   *                    _access_key=privake_key, 
   *                    _legacy_support=false);
   * \endcode
   * \return Conexo criada. A propriedade(ownership) da conexo  do usurio.
   *         A biblioteca no armazena internamente e de forma 
   *         permanente uma referncia forte para a conexo.
   *
   */
  BOOST_PARAMETER_MEMBER_FUNCTION(
    (boost::shared_ptr<Connection>),
    connectByReference,
    tag,
    (required
     (reference, (CORBA::Object_var)))
    (optional
     (access_key, (EVP_PKEY*), (EVP_PKEY*)0)
     (legacy_support, (bool), true))
  )
  {
    return connect_by_reference_impl(
      reference, access_key, legacy_support);
  }

  /**
   * \brief Cria uma conexo para um barramento indicado por um nome 
   *        ou endereo de rede.
   * 
   * O barramento  indicado por um nome ou endereo de rede e um
   * nmero de porta, onde os servios ncleo daquele barramento esto
   * executando. Este mtodo aceita argumentos no formato Boost Parameter.
   * 
   * \param _host Endereo ou nome de rede onde os servios ncleo do 
   *            barramento esto executando.
   * \param _port Porta onde os servios ncleo do barramento esto 
   *            executando.
   * \param _access_key (opcional) chave de acesso a ser utilizada
   *            internamente para a gerao de credenciais que identificam 
   *            as chamadas atravs do barramento. A chave deve ser uma 
   *            chave privada RSA de 2048 bits (256 bytes). O parmetro 
   *            do tipo EVP_PKEY* e na sua ausncia uma chave de acesso 
   *            gerada automaticamente.
   * \param _legacy_support (opcional) boleano que desabilita o suporte
   *            a chamadas usando o protocolo OpenBus 2.0. Por padro o suporte
   *            est habilitado
   *
   * Exemplo de chamadas sem os parmetros opcionais:
   * \code
   * connectByAddress("localhost", 2089);
   * connectByAddress(_host="localhost", _port=2089);
   * \endcode
   *
   * Exemplo de chamadas com os parmetros opcionais:
   * \code
   * connectByAddress("localhost", 2089, 
   *                  _access_key=privake_key, 
   *                  _legacy_support=false);
   *
   * connectByAddress("localhost", 2089, 
   *                  _legacy_support=true, 
   *                  _access_key=privake_key);
   * \endcode
	 * \throw InvalidBusAddress Os parmetros '_host' e '_port' no so vlidos.
   * \return Conexo criada. A propriedade(ownership) da conexo  do usurio.
   *         A biblioteca no armazena internamente e de forma 
   *         permanente uma referncia forte para a conexo.
   *
   */
  BOOST_PARAMETER_MEMBER_FUNCTION(
    (boost::shared_ptr<Connection>),
    connectByAddress,
    tag,
    (required
     (host, (const std::string&))
     (port, (unsigned short)))
    (optional
     (access_key, (EVP_PKEY*), (EVP_PKEY*)0)
     (legacy_support, (bool), true))
  )
  {
    return connect_by_address_impl(
      host, port, access_key, legacy_support);
  }
    
  /**
   * \brief Cria uma conexo para um barramento indicado por um nome 
   *        ou endereo de rede (deprecado).
   * 
   * \deprecated Ver \ref o mtodo \ref connectByAddress.
   *
   */
  BOOST_PARAMETER_MEMBER_FUNCTION(
    (boost::shared_ptr<Connection>), createConnection, tag,
    (required
     (host, (const std::string&))
     (port, (unsigned short)))
    (optional
     (access_key, (EVP_PKEY*), (EVP_PKEY*)0)
     (legacy_support, (bool), true))
  )
  {
    return connect_by_address_impl(
      host, port, access_key, legacy_support);
  }

  /**
   * \brief Define a conexo padro a ser usada nas chamadas.
   * 
   * Define uma conexo a ser utilizada em chamadas sempre que no houver uma
	 * conexo especfica definida no contexto atual, como  feito atravs da
	 * operao \ref setCurrentConnection. 
   * 
   * \param conn Conexo a ser definida como conexo padro. Um valor nulo 
   * significa nenhuma conexo definida como padro. OpenBusContext no
   * compartilha a propriedade sobre a conexo. Internamente referncias fortes
   * podem ser obtidas para essa conexo temporariamente durante o envio ou 
   * recebimento de chamadas remotas.
   */
  boost::shared_ptr<Connection>
    setDefaultConnection(const boost::shared_ptr<Connection> &);
   
  /**
   * \brief Devolve a conexo padro.
   * 
   * Veja operao \ref setDefaultConnection.
   * 
   * \return Conexo definida como conexo padro. \ref OpenBusContext no
   * compartilha a propriedade sobre a conexo. Internamente referncias fortes
   * podem ser obtidas para essa conexo temporariamente durante o envio ou 
   * recebimento de chamadas remotas.
   */
  boost::shared_ptr<Connection>
    getDefaultConnection() const;
   
  /**
   * \brief Define a conexo associada ao contexto corrente.
   * 
   * Define a conexo a ser utilizada em todas as chamadas
   * feitas no contexto atual, i.e., por exemplo uma thread de execuo.
   * 
   * \param conn Conexo a ser associada ao contexto corrente. 
   * Um valor nulo significa nenhuma conexo definida. 
   * Internamente referncias fortes
   * podem ser obtidas para essa conexo temporariamente durante o envio ou 
   * recebimento de chamadas remotas.
   */
  boost::shared_ptr<Connection>
    setCurrentConnection(const boost::shared_ptr<Connection> &);
   
  /**
   * \brief Devolve a conexo associada ao contexto corrente.
   * 
   * \return Conexo ao barramento associada ao contexto
   * corrente, que pode ter sido definida usando a operao 
   * \ref setCurrentConnection ou \ref setDefaultConnection. 
   * Internamente referncias fortes
   * podem ser obtidas para essa conexo temporariamente durante o envio ou 
   * recebimento de chamadas remotas.
   */
  boost::shared_ptr<Connection>
    getCurrentConnection() const;
       
  /**
   * \brief Devolve a cadeia de chamadas  qual a execuo corrente pertence.
   * 
   * Caso a contexto corrente (e.g. definido pelo 'CORBA::PICurrent')
   * seja o contexto de execuo de uma chamada remota oriunda do
   * barramento dessa conexo, essa operao devolve um objeto que
   * representa a cadeia de chamadas do barramento que esta chamada
   * faz parte. Caso contrrio, devolve uma cadeia de chamadas
   * 'vazia', representada por um CallerChain 'default-constructed'.
   *
   * Para verificar se a cadeia retornada  vlida, o seguinte idioma
   *  usado:
   *
   * \code
   * CallerChain chain(context.getCallerChain())
   * if(chain != CallerChain())
   *   // chain  vlido
   * else
   *   // chain  invlido
   * \endcode
   * \return Cadeia da chamada em execuo.
   */
  CallerChain getCallerChain();
  
  /**
   * \brief Associa uma cadeia de chamadas ao contexto corrente.
   * 
   * De forma que todas as chamadas remotas seguintes neste mesmo
   * contexto sejam feitas como parte dessa cadeia de chamadas.
   * 
   * \param chain Cadeia de chamadas a ser associada ao contexto
   * corrente. A ausncia de valor ou uma cadeia 'vazia' implica na
   * utilizao da cadeia obtida atravs de getCallerChain.
   *
   */
  void joinChain(const CallerChain &chain = CallerChain());
  
  /**
   * \brief Faz com que nenhuma cadeia de chamadas esteja associada ao
   * contexto corrente.
   * 
   * Remove a associao da cadeia de chamadas ao contexto corrente,
   * fazendo com que todas as chamadas seguintes feitas neste mesmo
   * contexto deixem de fazer parte da cadeia de chamadas associada
   * previamente. Ou seja, todas as chamadas passam a iniciar novas
   * cadeias de chamada.
   */
  void exitChain();

  /**
   * \brief Devolve a cadeia de chamadas associada ao contexto corrente.
   * 
   * A cadeia de chamadas
   * informada foi associada previamente pela operao
   * \ref joinChain. Caso o contexto corrente no tenha nenhuma cadeia
   * associada, essa operao devolve uma cadeia 'vazia'
   * 'default-constructed'
   * 
   * Para verificar se a cadeia retornada  vlida, o seguinte idioma  usado:
   * \code
   * CallerChain chain(context.getCallerChain())
   * if(chain != CallerChain())
   *   // chain  vlido
   * else
   *   // chain  invlido
   * \endcode
   * \return Cadeia de chamadas associada ao contexto corrente ou uma
   * cadeia 'vazia'.
   *
   */
  CallerChain getJoinedChain() const;

  /**
   * \brief Cria uma cadeia de chamadas para a entidade especificada.
   * 
   * O dono da cadeia  a conexo corrente \ref getCurrentConnection()
   * e utiliza-se a cadeia atual \ref getJoinedChain() como a cadeia
   * que se deseja dar seguimento ao encadeamento.  permitido
   * especificar qualquer nome de entidade, tendo ela um login ativo
   * no momento ou no.  A cadeia resultante s conseguir ser
   * utilizada \ref joinChain() com sucesso por uma conexo que possua
   * a mesma identidade da entidade especificada.
   *
   * Para verificar se a cadeia retornada  vlida, o seguinte idioma  usado:
   * \code
   * CallerChain chain(context.makeChainFor(loginId));
   * if(chain != CallerChain())
   *   // chain  vlido
   * else
   *   // chain  invlido
   * \endcode
   * \param entity nome da entidade para a qual deseja-se enviar a
   *        cadeia.
   * \throw CORBA::Exception
   * \return a cadeia gerada para ser utilizada pela entidade com o login
   *         especificado.
   */
  CallerChain makeChainFor(const std::string &entity) const;

	/**
	 * \brief Cria uma cadeia de chamadas assinada pelo barramento com
	 *        informaes de uma autenticao externa ao barramento.
	 * 
	 * A cadeia criada pode ser usada pela entidade do login que faz a chamada.
	 * O contedo da cadeia  dado pelas informaes obtidas atravs do token
	 * indicado.
	 * 
	 * \param token Valor opaco que representa uma informao de autenticao 
   *        externa.
	 * \param domain Identificador do domnio de autenticao.
	 * \return A nova cadeia de chamadas assinada.
	 *
	 * \exception idl::access::InvalidToken O token fornecido no foi reconhecido.
	 * \exception idl::access::UnknownDomain O domnio de autenticao no  
   *            conhecido.
	 * \exception idl::access::WrongEncoding A importao falhou, pois o token no 
   *            foi codificado corretamente com a chave pblica do barramento.
	 * \exception idl::core::ServiceFailure Ocorreu uma falha interna nos servios 
   *            do barramento que impediu a criao da cadeia.
	 */
	CallerChain importChain(
    const CORBA::OctetSeq &token,
    const std::string &domain = "") const;

  /**
   * \brief Codifica uma cadeia de chamadas (CallerChain) para um stream de
   * bytes.
   * 
   * Codifica uma cadeia de chamadas em um stream de bytes para permitir a
   * persistncia ou transferncia da informao. 
   * 
   * \param chain A cadeia a ser codificada.
   * \return A cadeia codificada em um stream de bytes.
   */
  CORBA::OctetSeq encodeChain(const CallerChain chain);

  /**
   * \brief Decodifica um stream de bytes de uma cadeia para o formato
   * CallerChain.
   * 
   * \param encoded O stream de bytes que representa a cadeia.
   * \throw InvalidEncodedStream Caso o stream de bytes no seja do
   *        formato esperado.
   * \return A cadeia de chamadas no formato CallerChain.
   */
  CallerChain decodeChain(const CORBA::OctetSeq &encoded) const;
  
  /**
	 * \brief Codifica um segredo de autenticao compartilhada
   *        SharedAuthSecret para um stream de bytes.
   * 
   * A codificao permite a persistncia ou transferncia da informao.
   * 
   * \param secret Segredo de autenticao compartilhada a ser codificado.
   * \return Cadeia codificada em um stream de bytes.
   */
  CORBA::OctetSeq encodeSharedAuthSecret(const SharedAuthSecret &secret);

  /**
   * \brief Decodifica um segredo de autenticao compartilhada
   *				SharedAuthSecret a partir de um stream de bytes.
   * 
   * \param encoded Stream de bytes contendo a codificao do segredo.
   * \throw InvalidEncodedStream Caso a stream de bytes no seja do formato
   *						 esperado.
   * \return Segredo de autenticao compartilhada decodificado.
   */
  SharedAuthSecret decodeSharedAuthSecret(const CORBA::OctetSeq &encoded);

  /** 
   * \brief Referncia ao ORB controlado pelo contexto.
   */
  CORBA::ORB_ptr orb() const 
  {
    return _orb;
  }
  
	/**
	 * \brief Referncia ao servio ncleo de registro de ofertas do barramento
	 *        referenciado no contexto atual.
	 */
  idl::offers::OfferRegistry_ptr getOfferRegistry() const;
  
	/**
	 * \brief Referncia ao servio ncleo de registro de logins do barramento
	 *        referenciado no contexto atual.
	 */
  idl::access::LoginRegistry_ptr getLoginRegistry() const;
  
  ~OpenBusContext();
private:
  /**
   * OpenBusContext deve ser adquirido atraves de:
   *   orb->resolve_initial_references("OpenBusContext")
   */
  OpenBusContext(CORBA::ORB_ptr, interceptors::ORBInitializer *);
  
  OpenBusContext(const OpenBusContext&);
  OpenBusContext &operator=(const OpenBusContext &);

  void orb(CORBA::ORB_ptr o) 
  {
    _orb = o;
  }

  CORBA::OctetSeq encode_exported_versions(idl::data_export::VersionedDataSeq,
                                           const std::string &tag);

  std::string decode_exported_versions(
    const CORBA::OctetSeq &stream,
    idl::data_export::VersionedDataSeq_out exported_version_seq) const;

  boost::shared_ptr<Connection> connect_by_address_impl(
    const std::string &host,
    unsigned short port,
    EVP_PKEY *access_key,
    bool legacy_support);

  boost::shared_ptr<Connection> connect_by_reference_impl(
    CORBA::Object_var,
    EVP_PKEY *access_key,
    bool legacy_support);

  boost::shared_ptr<Connection> create_connection_impl(
    const std::string &host,
    unsigned short port,
    EVP_PKEY *access_key,
    bool legacy_support);

  CallerChain extract_call_chain(
    idl::creden::SignedData,
    const boost::shared_ptr<Connection> &);
  
  CallerChain extract_legacy_call_chain(
    idl::legacy::creden::SignedCallChain,
    const boost::shared_ptr<Connection> &);

  mutable boost::mutex _mutex;
  interceptors::ORBInitializer * _orb_init;
  CORBA::ORB_ptr _orb;
  boost::weak_ptr<Connection> _def_conn;
  CallDispatchCallback _call_dispatch_cbk;
  std::map<boost::uuids::uuid, boost::weak_ptr<Connection> > id2conn;
  std::set<boost::weak_ptr<Connection> > connections;

  friend std::auto_ptr<orb_ctx> openbus::ORBInitializer(
    int &argc, char **argv);
  friend struct interceptors::ServerInterceptor;
  friend struct interceptors::ClientInterceptor;
};
}

#endif

